#include "fitline.h"

// allocates memory for n-component vector of doubles

dbl* dvector(int n)
{
    dbl* v;
    v = (dbl*) calloc((size_t)n, sizeof(dbl));
    if ( !v ) {
        err("vector allocation failure");
    }
    return v ;
}

// al kelley's graceful file closer


void gfclose(FILE* fp)
{
   if (fclose(fp) == EOF) {
      err("fclose() failed");
   }
}
// al kelley's graceful file opener


FILE* gfopen(char* filename, char* mode)
{
   FILE* fp;

   if ((fp = fopen(filename, mode)) == 0) {
      err("could not open file ", filename);
   }
   return fp;
}

//
// returns the logarithm of n factorial.
// from numerical recipes 2/e page 215.
// renamed to make it easier to recall.
// 5/16/97
// 


float lnfac( int n )
{
    static float a[101] ;  /* a static array initialized to zero */

    assert( n >= 0 ) ;
    if ( n <= 1 ) return 0.0 ;
    if ( n <= 100 )        /* in range of table */
        return a[n] ? a[n] : ( a[n] = lngam( n + 1.0 )) ;
    else                   /* out of range of table */
        return lngam( n + 1.0 ) ;
}

//
//
// returns incomplete gamma function       *
// Q(a,x) = 1 - P(a,x)               *
// this is a modification of the code given   *
// by press 2/e page 218, to make convergence *
// better. converges for 0.0 < a <= 2624 now. *
// written 5/18/97                            *
//

float gammq(float a, float x)
{
    float gamser,  /* for the series rep */
          gamcf,   /* for the continued fraction rep */
          lng ;    /* log of gamma(a) */

    if ( x < 0.0 || a <= 0.0 )
        err("invalid arguments in ", "gammq");
    if ( x > 3.0 ) {
        if ( x <= a - 1.0 ) {
            gser( &gamser, a, x, &lng ) ;
            return 1.0 - gamser ;
        }
        else {     /* x == a - 1.0 not allowed in gcf */
            gcf( &gamcf, a, x, &lng ) ;
            return gamcf ;
        }
    }
    else {  /* 0.0 <= x <= 3.0 */
        if ( x < a + 1.0 ) {
            gser( &gamser, a, x, &lng ) ;
            return 1.0 - gamser ;
        }
        else {
            gcf( &gamcf, a, x, &lng ) ;
            return gamcf ;
        }
    }
}

// returns incomplete gamma function P(a,x),
// evaluated by its series rep, as gamser.
// also returns ln gamma(a) as lng. from press 2/e p218

void gser(float* gamser, float a, float x, float* lng)
{
    int n;
    float sum, del, ap;

    *lng = lngam(a);
    if ( x <= 0.0 ) {
        if (x < 0.0)
            err("x less than 0.0 in routine gser");
        *gamser = 0.0;
        return;
    } else {
        ap = a;
        del = sum = 1.0/a;
        for ( n = 1 ; n <= itmax ; n++ ) {
            ++ap;
            del *= x/ap;
            sum += del;
            if (fabs( del ) < fabs( sum )*eps) {
                assert( x > 0.0 );
                *gamser = sum * exp( -x + a*log(x) - (*lng));
                return;
            }
        }
        fprintf( stderr, "\nat x = %7.2f a = %7.2f\n", x, a );
        err("\"a\" (nu/2) too large "
                "or itmax too small in routine gser");
        return;
    }
}

//
// returns incomplete gamma function Q(a,x) = 1.0-P(a,x),
// evaluated by its continued fraction rep, as gamcf.
// also returns ln gamma(a) as lng. from press 2/e p218
//


const float FPMIN = 1.0e-30; // near the smallest representable float

void gcf(float* gamcf, float a, float x, float* lng)
{
    int i;
    float an, b, c, d, del, h;

    *lng = lngam( a );
    b = x + 1.0 - a;
    c = 1.0 / FPMIN;
	assert( b != 0.0 );
    d = 1.0 / b;
    h = d;
    for (i = 1; i <= itmax ; i++) {
        an = -i * ( i - a );
        b += 2.0;
        d = an * d + b;
        if ( fabs( d ) < FPMIN ) d = FPMIN;
        c = b + an / c;
        if ( fabs( c ) < FPMIN ) c = FPMIN;
		assert( d != 0.0 );
        d = 1.0 / d;
        del = d * c;
        h *= del;
        if (fabs(del - 1.0) < eps)
            break;
    }
    if (i > itmax) {
		printf("near x = %f\n", x ) ;
        err("\"a\" too large, itmax too small in routine gcf");
	}
	assert(x > 0.0);
    *gamcf = exp( -x + a * log( x ) - ( *lng )) * h;
}


// returns ln gamma(xx) for xx > 0
// see press 2/e p213, who uses gammln for the name.


float lngam(float xx)
{
    dbl x, y, tmp, ser;
    static dbl cof[6] = { 76.18009172947146, -86.50532032941677,
        24.01409824083091, -1.231739572450155,
        0.1208650973866179e-2, -0.5395239384953e-5 };
    int j;

    assert(xx > 0.0);
    y = x = xx;
    tmp = x + 5.5;
	assert(tmp > 0.0);
    tmp -= ( x + 0.5 ) * log( tmp );
    ser = 1.000000000190015;
    for ( j = 0 ; j < 6 ; j++ )
        ser += cof[j]/++y ;
    return -tmp + log( 2.5066282746310005 * ser/x );
}
